/*
 * Copyright (c) 2004,2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "evilwm.h"

/**
 ** Create a Unix domain socket and listen for connection requests 
 **/
int
create_socket(int port)
{
	int fd;
	struct sockaddr_in sa;
	int reuse = 1;
	struct linger lg;
	struct protoent *pe;

	pe = getprotobyname("tcp");
	if (pe == NULL) {
		perror("getprotoent");
		return -1;
	}
	
	fd = socket(PF_INET, SOCK_STREAM, pe->p_proto);
	if (fd < 0) {
		perror("socket");
		return -1;
	}
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse,
		sizeof(reuse)) < 0) {
		perror("setsockopt/REUSEADDR");
		return -1;
	}
	lg.l_onoff = 0;
	lg.l_linger = 0;
	if (setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *)&lg,
		sizeof(lg)) < 0) {
		perror("setsockopt/LINGER");
		return -1;
	}
	memset(&sa, 0, sizeof(sa));

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(port);
	
	if (bind(fd, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
		fprintf(stderr, "bind: %d: %s\n", port,
			strerror(errno));
		close(fd);
		return -1;
	}
	if (listen(fd, 5) < 0) {
		perror("listen");
		close(fd);
		return -1;
	}
	return fd;
}

/**
 ** This function is called by the X Toolkit (in XtNextEvent()) 
 ** whenever a client connects to the socket
 **/
void
connect_client(XtPointer client_data, int *source, XtInputId *id)
{
	int s;
	socklen_t len;
	struct sockaddr_in sa;
	FILE *f;

	len = sizeof(sa);
	s = accept(*source, (struct sockaddr *)&sa, &len);
	if (s < 0) {
		perror("accept");
		close(*source);
		return;
	}
#ifdef DEBUG
	fprintf(stderr, "connection on socket\n");
#endif
	/* Open a stream over the file descriptor */
	f = fdopen(s, "r+");
	/* Register the callback to read actual data with Xt */
	XtAppAddInput((XtAppContext)client_data, s,
		      (XtPointer)XtInputReadMask, handle_cmd, 
		      (XtPointer)f);
}

#define MAXTOKENS 128

/**
 ** This function is called by the X Toolkit (in XtNextEvent()) 
 ** whenever data is available to read on the socket
 **/
void
handle_cmd(XtPointer client_data, int *source, XtInputId *id)
{
	char cmd[1024], *p;
	const char *argv[MAXTOKENS];
	char *last;
	const char *result;
	int argc = 0;
	FILE *f = (FILE *)client_data;
	Client *c;

	if (fgets(cmd, sizeof(cmd), f) == NULL) {
#ifdef DEBUG
	  fprintf(stderr, "client deconnection\n");
#endif
		XtRemoveInput(*id);
		fclose(f);
		/* handle_signal(0, NULL); */
		return;
	}
#ifdef DEBUG
	fprintf(stderr, "-> %s\n", cmd);
#endif
	/* split command into words */
	for ((p = strtok_r(cmd, " \r\n", &last)); p; 
	     (p = strtok_r(NULL, " \r\n", &last))) {
		if (argc < MAXTOKENS - 1) 
			argv[argc++] = p;
	}
	argv[argc] = NULL;
	
	result = "OK";

	/* parse commands */
	if (strncmp(argv[0], "LIST", 4) == 0) {
		dump_clients(f);
	} else if (strncmp(argv[0], "NEXT", 4) == 0) {
		next();
	} else if (strncmp(argv[0], "EXIT", 4) == 0) {
		fclose(f);
		handle_signal(0, NULL);
	} else if (strncmp(argv[0], "SELE", 4) == 0) {
		c = find_client_by_name(argv[1]);
		if (c != NULL) {
			if (vdesk != c->vdesk) 
				switch_vdesk(c->vdesk);
			unhide(c, RAISE);
			setmouse(c->window, 0, 0);
			setmouse(c->window, 
				 c->width + c->border - 1,
				 c->height + c->border - 1);
			client_update_current(current, c);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "SCREEN", 3) == 0) {
		XWindowAttributes xwa;
		XGetWindowAttributes(dpy, XDefaultRootWindow(dpy), &xwa);
		fprintf(f, "SCREEN %d %d\n", xwa.width, xwa.height);
	} else if (strncmp(argv[0], "NEW", 3) == 0) {
		if (argc == 1) {
			spawn(opt_term);
		} else {
			argv[0] = argv[1];
			spawn(argv);
		}
	} else if (strncmp(argv[0], "KILL", 4) == 0) {
		c = find_client_by_name(argv[1]);
		if (c != NULL) {
			send_wm_delete(c);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "MOVE", 4) == 0) {
		c = find_client_by_name(argv[1]);
		if (c!= NULL) {
			c->x = atoi(argv[2]);
			c->y = atoi(argv[3]);
			move(c, 1);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "GEOM", 4) == 0) {
		c = find_client_by_name(argv[1]);
		if (c != NULL) {
			c->x = atoi(argv[2]);
			c->y = atoi(argv[3]);
			c->width = atoi(argv[4]);
			c->height = atoi(argv[5]);
#ifdef DEBUG
			fprintf(stderr, "new geom %d %d %dx%d\n",
			    c->x, c->y, 
			    c->width, c->height);
#endif
			move(c, 1);
			resize(c, 1);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "HIDE", 4) == 0) {
		c = find_client_by_name(argv[1]);
		if (c != NULL) {
			hide(c);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "SHOW", 4) == 0) {
		c = find_client_by_name(argv[1]);
		if (c != NULL) {
			unhide(c, 1);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "LOW", 3) == 0) {
		c = find_client_by_name(argv[1]);
		if (c != NULL) {
			XLowerWindow(dpy, c->parent);
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "VDESK", 5) == 0) {
		int d = atoi(argv[1]);
#ifdef DEBUG
		fprintf(stderr, "new vdesk %d\n", d);
#endif
		switch_vdesk(d);
	} else if (strncmp(argv[0], "TODESK", 6) == 0) {
		int d = atoi(argv[2]);
		c = find_client_by_name(argv[1]);
#ifdef DEBUG
		fprintf(stderr, "to vdesk %d\n", d);
#endif
		if (c != NULL) {
			c->vdesk = d;
			if (vdesk != d) {
				hide(c);
			} else {
				unhide(c, NO_RAISE);
			}
		} else {
			result = "ERROR";
		}
	} else if (strncmp(argv[0], "TRANS", 5) == 0) {
		c = find_client_by_name(argv[1]);
		if (c == NULL) 
			result = "ERROR";
		else {
			double t = strtod(argv[2], NULL);
			settrans(c, t);
		}
	} else if (strcmp(argv[0], "FADE") == 0) {
		c = find_client_by_name(argv[1]);
		if (c == NULL)
			result = "ERROR";
		else 
			fade(c, atoi(argv[2]));
	} else if (strncmp(argv[0], "FADE2", 5) == 0) {
		Client *c2;
		c = find_client_by_name(argv[1]);
		c2 = find_client_by_name(argv[2]);
		if (c == NULL || c2 == NULL)
			result = "ERROR WINDOW NOT FOUND";
		else
			fade2(c, c2);
	} else if (strcmp(argv[0], "GAMMA") == 0) {
		if (set_gamma(dpy, current_screen->screen, strtod(argv[1], NULL)) < 0)
			result = "ERROR";
	} else {
		/* command unknown */
		result = "ERROR CMD NOT FOUND";
	}
	fprintf(f, "%s\n", result);
	fflush(f);
}

